/*
 * (C) Copyright IBM Corp. 1999-2000  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package xss4jsamples.asn1;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;

import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Reader;
import java.io.Writer;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

import javax.swing.text.BadLocationException;

/**
  * DocPanel is the panel for documents.
  */
public class DocPanel extends JPanel {
    private JLabel openFileLabel; // a name of a opening file
    private JTextArea docTextArea; // a text area for documents
    private StringWriter docWriter; // a writer for documents

    /**
      * Constructs a panel for documents.
      */
    public DocPanel() {
        super();
        setLayout(new BorderLayout());
        setBorder(new TitledBorder(new EtchedBorder()," Doc Window "));

        JPanel openFilePanel = new JPanel();
        add(openFilePanel, BorderLayout.NORTH);
        openFilePanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        
        openFilePanel.add(new JLabel("File : "));
        openFileLabel = new JLabel("");
        openFileLabel.setForeground(Color.black);
        openFilePanel.add(openFileLabel);
        
        docTextArea = new JTextArea();
        add(new JScrollPane(docTextArea, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED), BorderLayout.CENTER);
    }

    /**
      * Appends the given text to the end of the document.
      *
      * @param <CODE>str</CODE> - the text to insert
      */
    public void append(String str) {
        docTextArea.append(str + "\n");
    }
    
    /**
      * Clears the contents in text area.
      *
      * @param <CODE>str</CODE> - the text to insert
      */
    public void clear() {
        docTextArea.setText("");
    }

    /**
      * Transfers the currently selected range in the associated text area to the system clipboard,
      * leaving the contents in the text area.
      */
    public void copy() {
        docTextArea.copy();
    }

    /**
      * Transfers the currently selected range in the associated text area to the system clipboard,
      * removing the contents from the text area. 
      */
    public void cut() {
        docTextArea.cut();
    }
    
    /**
      * Flush the stream. If the stream has saved any characters from the various write() methods
      * in a buffer, write them immediately to the text area.
      *
      * @exception java.io.IOException -
      */
    public void flush() throws IOException {
        if (docWriter == null) throw new IOException();

        read(new StringReader(docWriter.toString()), null);
    }

    /**
      * Returns the name of the opening file.
      *
      * @return the name of the opening file.
      */
    public String getFileName() {
        return openFileLabel.getText();
    }

    /**
      * Returns the contents in the text area as a <CODE>Reader</CODE> object.
      *
      * @return a <CODE>Reader</CODE> object containing the contents in the text area.
      */
    public Reader getReader() {
        return (Reader)new StringReader(getText());
    }

    /**
      * Returns the text contained in the text area.
      *
      * @return the text.
      */
    public String getText() {
        return docTextArea.getText();
    }

    /**
      * Fetches a portion of the text represented by the text area.
      *
      * @param <CODE>offs</CODE> - the offset >= 0.
      * @param <CODE>offs</CODE> - the length >= 0.
      *
      * @return the text.
      */
    public String getText(int offs, int len) throws BadLocationException {
        return docTextArea.getText(offs, len);
    }

    /**
      * Returns a <CODE>Writer</CODE> object that you can use to write character text
      * in the text area.
      *
      * @return a <CODE>Writer</CODE> object that can write text in the text area.
      */
    public Writer getWriter() {
        docWriter = new StringWriter();

        return (Writer)docWriter;
    }

    /**
      * Inserts the specified text at the specified position.
      *
      * @param <CODE>str</CODE> - the text to insert.
      * @param <CODE>pos</CODE> - the position at which to insert >= 0.
      */
    public void insert(String str, int pos) {
        docTextArea.insert(str, pos);
    }
    
    /**
      * Transfers the contents of the system clipboard into the text area.
      */
    public void paste() {
        docTextArea.paste();
    }

    /**
      * Initializes the text area from a stream. By default this will load the text area 
      * as plain text. Previous contents of the text area are discarded.
      *
      * @param <CODE>in</CODE> - The stream to read from
      * @param <CODE>desc</CODE> - An object describing the stream. This might be a string, a File, a URL, etc. 
      *
      * @exception java.io.IOException -
      */
    public void read(Reader in, Object desc) throws IOException {
        docTextArea.read(in, desc);
    }

    /**
      * Selects all the text in the text area.
      */
    public void selectAll() {
        docTextArea.selectAll();
    }

    /**
      * Sets the name of the opening file.
      *
      * @param <CODE>file</CODE> - the name of the opening file.
      */
    public void setFileName(String file) {
        openFileLabel.setText(file);
    }

    /**
      * Sets the contents of the text area to the specified text.
      *
      * @param <CODE>t</CODE> - the new text to be set.
      */
    public void setText(String t) {
        docTextArea.setText(t);
    }

    /**
      * Stores the contents of the text area into the given stream. 
      * By default this will store the text area as plain text.
      *
      * @param <CODE>out</CODE> - the output stream.
      *
      * @exception java.io.IOException -
      */
    public void write(Writer out) throws IOException {
        docTextArea.write(out);
    }
}
