/*
 * (C) Copyright IBM Corp. 1999-2000  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package xss4jsamples.asn1;

import com.ibm.xml.asn1.LabelSAX;

import com.ibm.xml.parser.DTD;
import com.ibm.xml.parser.Parser;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import java.net.URL;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
  * This class provides methods to translate a DER encoded ASN.1 PKCS#7 data into 
  * a labeled XML document using the "cascading" technique. 
  */
public class MyLabelPKCS7 extends DefaultHandler {
    private String rootName;
    private String dtdFile;
    private boolean isBuiltin;  // a flag to show whether a DTD is built-in

    private StringBuffer buffer = null;
    private String contentType;
    private ContentHandler handler;
    private ContentHandler handlerSave;

    /**
      * Constructs an object and set a document handler to it.
      *
      * @param <CODE>handler</CODE> - a document handler.
      *
      * @exception java.io.IOException -
      */
    public MyLabelPKCS7(ContentHandler handler) throws IOException {
        this.handler = handler;
    }

    /**
      * Sets the specification of a doctype declaration.
      *
      * @param <CODE>root</CODE> - the name of the root.
      * @param <CODE>file</CODE> - the name of the file containing the DTD.
      */
    public void setDoctypeDeclaration(String root, String file) {
      setDoctypeDeclaration(root, file, true);
    }

    public void setDoctypeDeclaration(String root, String file, boolean isBltn) {
      rootName = root;
      dtdFile = file;
      isBuiltin = isBltn;
    }

    /**
      * Receives notification of the beginning of a document.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void startDocument() throws SAXException {
        handler.startDocument();
    }

    /**
      * Receives notification of the end of a document.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void endDocument() throws SAXException {
        handler.endDocument();
    }

    /**
      * Receives notification of the beginning of an element.
      *
      * @param <code>uri</code> the Namespace URI.
      * @param <code>localName</code> the local name (without prefix).
      * @param <code>name</code> the qualified name (with prefix).
      * @param <code>atts</code> the attributes attached to the element.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void startElement(String uri, String localName, String name, Attributes atts) throws SAXException {
        handler.startElement(uri, localName, name, atts);

        if (name.equals("ContentInfo.contentType")) {
            buffer = new StringBuffer();
        } else if (name.equals("ContentInfo.content")) {
            // Start processing Content
            //  1. Create another LabelSAX
            LabelSAX lsaxContents = new LabelSAX(handler);
            try{
                DTD dtd = readDTD(dtdFile, isBuiltin);
                lsaxContents.setDTD(dtd, oid2ElementName(contentType));
            } catch (IOException ioe) {
                throw new SAXException(ioe.toString());
            }

            //  2. Redirect SAX events
            handlerSave = handler;
            handler = lsaxContents;
        }
    }

    /**
      * Reads a DTD from a file.
      *
      * @param <CODE>file</CODE> - the name of the file containing a DTD.
      *
      * @return a DTD.
      *
      * @exception java.io.IOException -
      */
    private DTD readDTD(String file, boolean isBltn) throws IOException {
      Parser psr = new Parser(file);
      psr.setErrorHandler(this);

      InputStream in = null;
      if (isBltn) {
	in = (MyTranslator.class).getResourceAsStream(file);

      } else {
	if (file.startsWith("file://")
	    || file.startsWith("http://")
	    || file.startsWith("ftp://")) {
	  in = (new URL(file)).openStream();

	} else {
	  in = new FileInputStream(file);
	}
      }
      DTD dtd = psr.readDTDStream(in);

      return dtd;
    }

    /**
     * Converts a PKCS#7 OID or a PKCS#7 OID name to a element name.
     *
     * @param <CODE>oid</CODE> - a OID or an OID name
     */
    private String oid2ElementName(String oid) {
        if (oid.equals("1.2.840.113549.1.7.1") || oid.equals("data")) return "root..Data";
        else if (oid.equals("1.2.840.113549.1.7.2") || oid.equals("signedData")) return "root..SignedData";
        else if (oid.equals("1.2.840.113549.1.7.3") || oid.equals("envelopedData")) return "root..EnvelopedData";
        else if (oid.equals("1.2.840.113549.1.7.4") || oid.equals("signedAndEnvelopedData")) return "root..SignedAndEnvelopedData";
        else if (oid.equals("1.2.840.113549.1.7.5") || oid.equals("digestedData")) return "root..DigestedData";
        else if (oid.equals("1.2.840.113549.1.7.6") || oid.equals("encryptedData")) return "root..EncryptedData";
        else return "";
    }


    /**
      * Receives notification of the end of an element.
      *
      * @param <code>uri</code> the Namespace URI.
      * @param <code>localName</code> the local name (without prefix).
      * @param <code>name</code> the qualified name (with prefix).
      *
      * @exception org.xml.sax.SAXException -
      */
    public void endElement(String uri, String localName, String name) throws SAXException {
        if (name.equals("ContentInfo.contentType")) {
            contentType = this.buffer.toString();
            buffer = null;
        } else if (name.equals("ContentInfo.content")) {
            // Restore handler
            handler = handlerSave;
        }

        handler.endElement(uri, localName, name);
    }

    /**
      * Receives notification of character data. 
      *
      * @param <CODE>chars</CODE> - characters from a document.
      * @param <CODE>start</CODE> - the start position in the array.
      * @param <CODE>length</CODE> - the number of characters to be read from the array.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void characters(char chars[], int start, int length) throws SAXException {
        handler.characters(chars,start,length);

        if (buffer != null) {
            buffer.append(chars, start, length);
        }
    }

    /**
      * Receives notification of a warning. The exception is thrown again.
      *
      * @param <CODE>saxpe</CODE> - the warning information encapsulated in a SAX parse exception.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void warning(SAXParseException saxpe) throws SAXException {
        throw saxpe;
    }

    /**
      * Receives notification of a recoverable error. The exception is thrown again.
      *
      * @param <CODE>exc</CODE> - the error information encapsulated in a SAX parse exception.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void error(SAXParseException saxpe) throws SAXException {
        throw saxpe;
    }

    /**
      * Receives notification of a non-recoverable error. The exception is thrown again.
      *
      * @param <CODE>saxpe</CODE> - the error information encapsulated in a SAX parse exception.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void fatalError(SAXParseException saxpe) throws SAXException {
        throw saxpe;
    }
}
