/*
 * (C) Copyright IBM Corp. 1999-2000  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package xss4jsamples.asn1;

import com.ibm.xml.asn1.ASN12SAX;
import com.ibm.xml.asn1.ASN1Syntax2DTD;
import com.ibm.xml.asn1.LabelSAX;
import com.ibm.xml.asn1.ParseException;
import com.ibm.xml.asn1.SAX2ASN1;
import com.ibm.xml.asn1.SAXEventDumper;
import com.ibm.xml.asn1.UnlabelSAX;

import com.ibm.xml.parser.DTD;
import com.ibm.xml.parser.Parser;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;

import java.net.URL;

import org.apache.xerces.parsers.SAXParser;

import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
  * This class enables the following:
  *
  * - Translation from DER encoded ASN.1 data to a plain XML document or to a labeled one
  *
  * - Translation from a labeled XML document to a plain one or to DER encoded ASN.1 data
  */
public class MyTranslator implements ErrorHandler {
    private Object src; // a source
    private String rootName; // a name of a root
    private String dtdFile; // a name of a file containing a DTD
    private boolean isBuiltin;  // a flag to show whether a DTD is built-in
    private boolean namedOID; // a flag for a named OID

    /**
      * Constructs an empty object.
      */
    public MyTranslator() {
        src = null;
        rootName = null;
        dtdFile = null;
	isBuiltin = true;
        namedOID = true;
    }

    /**
      * Sets a source <CODE>InputStream</CODE> object.
      *
      * @param <CODE>src</CODE> - a source <CODE>InputStream</CODE> object.
      */
    public void setInputStream(InputStream src) {
        this.src = src;
    }

    /**
      * Sets a source <CODE>Reader</CODE> object.
      *
      * @param <CODE>src</CODE> - a source <CODE>Reader</CODE> object.
      */
    public void setReader(Reader src) {
        this.src = src;
    }


    /**
      * Sets the specification of a doctype declaration.
      *
      * @param <CODE>root</CODE> - the name of the root.
      * @param <CODE>file</CODE> - the name of the file containing the DTD.
      */
    public void setDoctypeDeclaration(String root, String file) {
      setDoctypeDeclaration(root, file, true);
    }

    public void setDoctypeDeclaration(String root, String file, boolean isBltn) {
      rootName = root;
      dtdFile = file;
      isBuiltin = isBltn;
    }

    /**
      * Enables or disables the named OID.
      *
      * @param <CODE>b</CODE> - true to enable the named OID, false to disable.
      */
    public void setNamedOID(boolean b) {
        namedOID = b;
    }

    /**
      * Translates DER encoded ASN.1 data into a plain XML document.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>OutputStream</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startAS(OutputStream dst) throws IOException, SAXException {
        if ((src == null) || (dst == null)) throw new IOException();

        SAXEventDumper saxed = new SAXEventDumper();
        saxed.setWriter(new OutputStreamWriter(dst));
        saxed.setXMLDeclaration("1.0", null);

        ASN12SAX a2s;
        if (namedOID) {
            NamedOID noid = new NamedOID(saxed);
            a2s = new ASN12SAX(noid);
        } else {
            a2s = new ASN12SAX(saxed);
        }

        if (src instanceof InputStream) a2s.setInputStream((InputStream)src);
        else throw new IOException();
        a2s.setEncoding("base64");
        a2s.start();
    }

    /**
      * Translates DER encoded ASN.1 data into a plain XML document.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>Writer</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startAS(Writer dst) throws IOException, SAXException {
        if ((src == null) || (dst == null)) throw new IOException();

        SAXEventDumper saxed = new SAXEventDumper();
        saxed.setWriter(dst);
        saxed.setXMLDeclaration("1.0", null);

        ASN12SAX a2s;
        if (namedOID) {
            NamedOID noid = new NamedOID(saxed);
            a2s = new ASN12SAX(noid);
        } else {
            a2s = new ASN12SAX(saxed);
        }

        if (src instanceof InputStream) a2s.setInputStream((InputStream)src);
        else throw new IOException();
        a2s.setEncoding("base64");
        a2s.start();
    }

    /**
      * Translates DER encoded ASN.1 data into an XML document and labels the document.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>OutputStream</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startASS(OutputStream dst) throws IOException, SAXException {
        if ((src == null) || (dst == null) || (rootName == null) || (dtdFile == null)) throw new IOException();

        SAXEventDumper saxed = new SAXEventDumper();
        saxed.setWriter(new OutputStreamWriter(dst));
        saxed.setXMLDeclaration("1.0", null);
        saxed.setDoctypeDeclaration(rootName, dtdFile);

        LabelSAX lsax = new LabelSAX(saxed);
        DTD dtd = readDTD(dtdFile, isBuiltin);
        lsax.setDTD(dtd, rootName);
        lsax.setEncoding("base64");

        ASN12SAX a2s;
        if (namedOID) {
            NamedOID noid = new NamedOID(lsax);
            a2s = new ASN12SAX(noid);
        } else {
            a2s = new ASN12SAX(lsax);
        }

        if (src instanceof InputStream) a2s.setInputStream((InputStream)src);
        else throw new IOException();
        a2s.start();
    }

    /**
      * Translates DER encoded ASN.1 data into an XML document and labels the document.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>Writer</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startASS(Writer dst) throws IOException, SAXException {
        if ((src == null) || (dst == null) || (rootName == null) || (dtdFile == null)) throw new IOException();

        SAXEventDumper saxed = new SAXEventDumper();
        saxed.setWriter(dst);
        saxed.setXMLDeclaration("1.0", null);
        saxed.setDoctypeDeclaration(rootName, dtdFile);

        LabelSAX lsax = new LabelSAX(saxed);
        DTD dtd = readDTD(dtdFile, isBuiltin);
        lsax.setDTD(dtd, rootName);
        lsax.setEncoding("base64");

        ASN12SAX a2s;
        if (namedOID) {
            NamedOID noid = new NamedOID(lsax);
            a2s = new ASN12SAX(noid);
        } else {
            a2s = new ASN12SAX(lsax);
        }

        if (src instanceof InputStream) a2s.setInputStream((InputStream)src);
        else throw new IOException();
        a2s.start();
    }

    /**
      * Translates DER encoded ASN.1 data into an XML document and labels the document using cascading LabelSAX.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>OutputStream</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startASSS(OutputStream dst) throws IOException, SAXException {
        if ((src == null) || (dst == null) || (rootName == null) || (dtdFile == null)) throw new IOException();

        SAXEventDumper saxed = new SAXEventDumper();
        saxed.setWriter(new OutputStreamWriter(dst));
        saxed.setXMLDeclaration("1.0", null);
        saxed.setDoctypeDeclaration(rootName, dtdFile);

        MyLabelPKCS7 lpkcs7 = new MyLabelPKCS7(saxed);
        lpkcs7.setDoctypeDeclaration(rootName, dtdFile, isBuiltin);

        LabelSAX lsax = new LabelSAX(lpkcs7);
        DTD dtd = readDTD(dtdFile, isBuiltin);
        lsax.setDTD(dtd, rootName);
        lsax.setEncoding("base64");

        ASN12SAX a2s;
        if (namedOID) {
            NamedOID noid = new NamedOID(lsax);
            a2s = new ASN12SAX(noid);
        } else {
            a2s = new ASN12SAX(lsax);
        }

        if (src instanceof InputStream) a2s.setInputStream((InputStream)src);
        else throw new IOException();
        a2s.start();
    }

    /**
      * Translates DER encoded ASN.1 data into an XML document and labels the document using cascading LabelSAX.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>Writer</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startASSS(Writer dst) throws IOException, SAXException {
        if ((src == null) || (dst == null) || (rootName == null) || (dtdFile == null)) throw new IOException();

        SAXEventDumper saxed = new SAXEventDumper();
        saxed.setWriter(dst);
        saxed.setXMLDeclaration("1.0", null);
        saxed.setDoctypeDeclaration(rootName, dtdFile);

        MyLabelPKCS7 lpkcs7 = new MyLabelPKCS7(saxed);
        lpkcs7.setDoctypeDeclaration(rootName, dtdFile);

        LabelSAX lsax = new LabelSAX(lpkcs7);
        DTD dtd = readDTD(dtdFile, isBuiltin);
        lsax.setDTD(dtd, rootName);
        lsax.setEncoding("base64");

        ASN12SAX a2s;
        if (namedOID) {
            NamedOID noid = new NamedOID(lsax);
            a2s = new ASN12SAX(noid);
        } else {
            a2s = new ASN12SAX(lsax);
        }

        if (src instanceof InputStream) a2s.setInputStream((InputStream)src);
        else throw new IOException();
        a2s.start();
    }

    /**
      * Reads a DTD from a file.
      *
      * @param <CODE>file</CODE> - the name of the file containing a DTD.
      *
      * @return a DTD.
      *
      * @exception java.io.IOException -
      */
    private DTD readDTD(String file, boolean isBltn) throws IOException {
      Parser psr = new Parser(file);
      psr.setErrorHandler(this);

      InputStream in = null;
      if (isBltn) {
	in = (Translator.class).getResourceAsStream(file);

      } else {
	if (file.startsWith("file://")
	    || file.startsWith("http://")
	    || file.startsWith("ftp://")) {
	  in = (new URL(file)).openStream();

	} else {
	  in = new FileInputStream(file);
	}
      }
      DTD dtd = psr.readDTDStream(in);

      return dtd;
    }

    /**
      * Unlabels a labeled XML document
      *
      * @param <CODE>dst</CODE> - a destination <CODE>OutputStream</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startSS(OutputStream dst) throws IOException, SAXException {
        if ((src == null) || (dst == null)) throw new IOException();

        SAXEventDumper saxed = new SAXEventDumper();
        saxed.setWriter(new OutputStreamWriter(dst));

        UnlabelSAX usax = new UnlabelSAX(saxed);
        usax.setEncoding("base64");

        SAXParser psr = new SAXParser();
        psr.setContentHandler(usax);
        psr.setErrorHandler(this);
	psr.setFeature("http://xml.org/sax/features/validation", true);
        if (src instanceof InputStream) psr.parse(new InputSource((InputStream)src));
        else if (src instanceof Reader) psr.parse(new InputSource((Reader)src));
        else throw new IOException();
    }

    /**
      * Unlabels a labeled XML document
      *
      * @param <CODE>dst</CODE> - a destination <CODE>Writer</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startSS(Writer dst) throws IOException, SAXException {
        if ((src == null) || (dst == null)) throw new IOException();

        SAXEventDumper saxed = new SAXEventDumper();
        saxed.setWriter(dst);

        UnlabelSAX usax = new UnlabelSAX(saxed);
        usax.setEncoding("base64");

        SAXParser psr = new SAXParser();
        psr.setContentHandler(usax);
        psr.setErrorHandler(this);
	psr.setFeature("http://xml.org/sax/features/validation", true);
        if (src instanceof InputStream) psr.parse(new InputSource((InputStream)src));
        else if (src instanceof Reader) psr.parse(new InputSource((Reader)src));
        else throw new IOException();
    }

    /**
      * Unlabels a labeled XML document and translates it into DER encoded ASN.1 data.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>OutputStream</CODE> object.
      *
      * @exception java.io.IOException -
      * @exception org.xml.sax.SAXException -
      */
    public void startSSA(OutputStream dst) throws IOException, SAXException {
        if ((src == null) || (dst == null)) throw new IOException();

        SAX2ASN1 s2a = new SAX2ASN1();
        s2a.setOutputStream(dst);

        UnnamedOID uoid = new UnnamedOID(s2a);
        UnlabelSAX usax = new UnlabelSAX(uoid);

        SAXParser psr = new SAXParser();
        psr.setContentHandler(usax);
        psr.setErrorHandler(this);
	psr.setFeature("http://xml.org/sax/features/validation", true);
        if (src instanceof InputStream) psr.parse(new InputSource((InputStream)src));
        else if (src instanceof Reader) psr.parse(new InputSource((Reader)src));
        else throw new IOException();
    }

    /**
      * Translates an ASN.1 abstract syntax definition into a DTD.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>OutputStream</CODE> object.
      *
      * @exception java.io.IOException -
      */
    public void startAD(OutputStream dst) throws IOException, ParseException {
        if ((src == null) || (dst == null)) throw new IOException();

        ASN1Syntax2DTD a2d = new ASN1Syntax2DTD();
        if (src instanceof InputStream) a2d.translate(new InputStreamReader((InputStream)src), new OutputStreamWriter(dst));
        else if (src instanceof Reader) a2d.translate((Reader)src, new OutputStreamWriter(dst));
        else throw new IOException();
    }

    /**
      * Translates an ASN.1 abstract syntax definition into a DTD.
      *
      * @param <CODE>dst</CODE> - a destination <CODE>Writer</CODE> object.
      *
      * @exception java.io.IOException -
      */
    public void startAD(Writer dst) throws IOException, ParseException {
        if ((src == null) || (dst == null)) throw new IOException();

        ASN1Syntax2DTD a2d = new ASN1Syntax2DTD();
        if (src instanceof InputStream) a2d.translate(new InputStreamReader((InputStream)src), dst);
        else if (src instanceof Reader) a2d.translate((Reader)src, dst);
        else throw new IOException();
    }

    /**
      * Receives notification of a warning. The exception is thrown again.
      *
      * @param <CODE>saxpe</CODE> - the warning information encapsulated in a SAX parse exception.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void warning(SAXParseException saxpe) throws SAXException {
        throw saxpe;
    }

    /**
      * Receives notification of a recoverable error. The exception is thrown again.
      *
      * @param <CODE>exc</CODE> - the error information encapsulated in a SAX parse exception.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void error(SAXParseException saxpe) throws SAXException {
        throw saxpe;
    }

    /**
      * Receives notification of a non-recoverable error. The exception is thrown again.
      *
      * @param <CODE>saxpe</CODE> - the error information encapsulated in a SAX parse exception.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void fatalError(SAXParseException saxpe) throws SAXException {
        throw saxpe;
    }
}
