/*
 * (C) Copyright IBM Corp. 1999-2000  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package xss4jsamples.asn1;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import java.util.Hashtable;
import java.util.StringTokenizer;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
  * This class provides methods to convert OIDs (Object Identifiers) from the dotted decimal 
  * notation to more readable notation. This uses an OID table which is generated from 
  * the data collected by Harald Tveit Alvestrand.
  */
public class NamedOID extends DefaultHandler {
    private static final String OID_ELEMENT_NAME = "_OBJECT_IDENTIFIER";
    private static final String OID_TABLE_FILE = "oidtable.dat";
    
    private static Hashtable oidTable;
    private StringBuffer buffer = null;
    private ContentHandler handler;

    /**
      * Constructs an object and set a document handler to it, and prepare a OID table.
      *
      * @param <CODE>handler</CODE> - a document handler.
      *
      * @exception java.io.IOException -
      */
    public NamedOID(ContentHandler handler) throws IOException {
        this.handler = handler;
        oidTable = new Hashtable();

	InputStream in = (NamedOID.class).getResourceAsStream(OID_TABLE_FILE);
	BufferedReader br = new BufferedReader(new InputStreamReader(in));

        String line;
        int lineno = 0;
        while ((line = br.readLine()) != null) {
            lineno++;
            StringTokenizer st = new StringTokenizer(line);
            if (!st.hasMoreTokens()) {
                new IOException("Error in line " + lineno + " in " + OID_TABLE_FILE);
            }

            String key = st.nextToken();
            if (!st.hasMoreTokens()) {
                new IOException("Error in line " + lineno + " in " + OID_TABLE_FILE);
            }

            String dash = st.nextToken();  // "-" to be ignored
            StringBuffer sb = new StringBuffer();
            while (st.hasMoreTokens()) {
                sb.append(st.nextToken());
                if (st.hasMoreTokens()) sb.append(" ");
            }

            oidTable.put(key, sb.toString());
        }
    }

    /**
      * Receives notification of the beginning of a document.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void startDocument() throws SAXException {
        handler.startDocument();
    }

    /**
      * Receives notification of the end of a document.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void endDocument() throws SAXException {
        handler.endDocument();
    }

    /**
      * Receives notification of the beginning of an element.
      *
      * @param <code>uri</code> the Namespace URI.
      * @param <code>localName</code> the local name (without prefix).
      * @param <code>name</code> the qualified name (with prefix).
      * @param <code>atts</code> the attributes attached to the element.
      *
      * @exception org.xml.sax.SAXException -
      */
    public void startElement(String uri, String localName, String name, Attributes atts) throws SAXException {
        handler.startElement(uri, localName, name, atts);

        if (name.equals("_OBJECT_IDENTIFIER")) {
            buffer = new StringBuffer();
        }
    }

    /**
      * Receives notification of the end of an element.
      *
      * @param <code>uri</code> the Namespace URI.
      * @param <code>localName</code> the local name (without prefix).
      * @param <code>name</code> the qualified name (with prefix).
      *
      * @exception org.xml.sax.SAXException -
      */
    public void endElement(String uri, String localName, String name) throws SAXException {
        if (name.equals("_OBJECT_IDENTIFIER")) {
            String oid = buffer.toString();
            String oidName = (String) oidTable.get(oid);

            if (oidName != null) {
                handler.characters(oidName.toCharArray(), 0, oidName.length());
            } else {
                handler.characters(oid.toCharArray(), 0, oid.length());
            }
            buffer = null;
        }

        handler.endElement(uri, localName, name);
    }

    /**
      * Receives notification of character data. 
      *
      * @param <CODE>chars</CODE> - characters from a document.
      * @param <CODE>start</CODE> - the start position in the array.
      * @param <CODE>length</CODE> - the number of characters to be read from the array.
      *
      * @exception org.xml.sax.SAXException -
      */
      public void characters(char[] chars, int start, int length) throws SAXException {
        if (buffer != null) {
            buffer.append(chars, start, length);
        } else {
            handler.characters(chars, start, length);
        }
    }
}
