/*
 * (C) Copyright IBM Corp. 1999-2000  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package xss4jsamples.asn1;

import com.ibm.xml.asn1.ParseException;

import java.awt.Dimension;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Rectangle;
import java.awt.Toolkit;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import java.util.StringTokenizer;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JSplitPane;

import org.xml.sax.SAXException;

/**
  * This is a demo program using ASN.1/XML Translator package. It is a simple XML editor 
  * with ASN.1/XML translation function has GUI(Graphical User Interface).
  *
  * This program enables you to:
  *
  * - open a labeled XML document, edit it, and save as it is or as a DER encoded ASN.1 data
  * - open a DER encoded ASN.1 data as a labeled XML document, edit it, and save as it is or as a DER encoded ASN.1 data
  * -open an ASN.1 abstract syntax definition, edit it, and save as it is or as a DTD
  */
public class XMLEditor extends JFrame implements ActionListener {
    private static final String LDAP_ROOT_NAME = "root..LDAPMessage";
    private static final String PKCS10_ROOT_NAME = "root..CertificationRequest";
    private static final String PKCS7_ROOT_NAME = "root..ContentInfo";
    private static final String SNMP_ROOT_NAME = "root..Message";
    private static final String X509_ROOT_NAME = "root..Certificate";
    private static final String LDAP_DTD = "ldap.dtd";
    private static final String PKCS10_DTD = "pkcs10.dtd";
    private static final String PKCS7_DTD = "pkcs7.dtd";
    private static final String SNMP_DTD = "snmp.dtd";
    private static final String X509_DTD = "x509.dtd";
    private static final String ABOUT_MESSAGE = "XML Editor with AXT Function\nCopyright(C) 2000 IBM Corp.\nAll Rights Reserved.";

    private JMenuItem openXMLMenuItem,openASN1WithoutLabelMenuItem,openASN1WithLabelExternalDTDMenuItem,openASN1WithLabelBuiltinDTDLDAPMenuItem,openASN1WithLabelBuiltinDTDPKCS10MenuItem,openASN1WithLabelBuiltinDTDPKCS7MenuItem,openASN1WithLabelBuiltinDTDSNMPMenuItem,openASN1WithLabelBuiltinDTDX509MenuItem,openASN1ASDMenuItem,saveMenuItem,saveAsXMLMenuItem,saveAsASN1MenuItem,saveAsASN1ASDMenuItem,saveAsDTDMenuItem,exitMenuItem;
    private JMenuItem cutMenuItem,copyMenuItem,pasteMenuItem,selectAllMenuItem;
    private JCheckBoxMenuItem oidMenuItem,cascadingPKCS7MenuItem;
    private JMenuItem aboutMenuItem;
    private JMenuItem exitPopupMenuItem,aboutPopupMenuItem;
    private JPopupMenu popupMenu;
    private JFileChooser fileChooser;

    private static SplashWindow splashWindow; // a window for a splash window
    private File openFile; // a opening file
    private DocPanel docPanel; // a panel for documents
    private LogPanel logPanel; // a panle for logs
    private MyTranslator tran; // a object for ASN,1/XML translation
    private ExternalDTDDialog dialog; // a dialog for using a external DTD

    /**
      * Constructs a main frame.
      *
      * @param <CODE>title</CODE> - the title for the frame.
      * @param <CODE>si</CODE> - the name of the file containing a splash image.
      */
    public XMLEditor(String title, String si) {
        super(title);

        // if specified the "-sw" option
        if (si != null) {
            MediaTracker mediaTracker = new MediaTracker(this);
            Image splashImage = Toolkit.getDefaultToolkit().getImage(si);
            mediaTracker.addImage(splashImage, 0);

            try{
                mediaTracker.waitForID(0);
            }
            catch(InterruptedException ie) {
                ie.printStackTrace();
            } 

            splashWindow = new SplashWindow(this, splashImage, splashImage.getWidth(null), splashImage.getHeight(null)); 
            splashWindow.draw();
        }

        setSize(800,600);
        Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
        Rectangle frameDim = getBounds();
        setLocation((screenDim.width - frameDim.width) / 2,(screenDim.height - frameDim.height) / 2); 

        tran = new MyTranslator();
        dialog = new ExternalDTDDialog(this, "ASN.1 File Open with Label Using External DTD", true);
        fileChooser = new JFileChooser();
        fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fileChooser.setFileHidingEnabled(false);
        fileChooser.setMultiSelectionEnabled(false);

        docPanel = new DocPanel();
        logPanel = new LogPanel();
        JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, true, docPanel, logPanel);
        splitPane.setDividerLocation(400);
        getContentPane().add(splitPane);

        JMenuBar menuBar = new JMenuBar();
        setJMenuBar(menuBar);

        JMenu fileMenu = new JMenu("File");
        JMenu editMenu = new JMenu("Edit");
        JMenu helpMenu = new JMenu("Help");
        menuBar.add(fileMenu);
        menuBar.add(editMenu);
        menuBar.add(helpMenu);

        JMenu openMenu = new JMenu("Open");
        saveMenuItem = new JMenuItem("Save");
        JMenu saveAsMenu = new JMenu("Save As");
        exitMenuItem = new JMenuItem("Exit");
        fileMenu.add(openMenu);
        fileMenu.add(saveMenuItem);
        fileMenu.add(saveAsMenu);
        fileMenu.addSeparator();
        fileMenu.add(exitMenuItem);
        saveMenuItem.addActionListener(this);
        exitMenuItem.addActionListener(this);

        openXMLMenuItem = new JMenuItem("XML");
        JMenu openASN1Menu = new JMenu("ASN.1");
        openASN1ASDMenuItem = new JMenuItem("ASN.1 ASD");
        openMenu.add(openXMLMenuItem);
        openMenu.add(openASN1Menu);
        openMenu.addSeparator();
        openMenu.add(openASN1ASDMenuItem);
        openXMLMenuItem.addActionListener(this);
        openASN1ASDMenuItem.addActionListener(this);

        JMenu openASN1WithLabelMenu = new JMenu("With Label");
        openASN1WithoutLabelMenuItem = new JMenuItem("Without Label");
        openASN1Menu.add(openASN1WithLabelMenu);
        openASN1Menu.add(openASN1WithoutLabelMenuItem);
        openASN1WithoutLabelMenuItem.addActionListener(this);

        JMenu openASN1WithLabelBuiltinDTDMenu = new JMenu("Using Built-in DTD");
        openASN1WithLabelExternalDTDMenuItem = new JMenuItem("Using External DTD");
        openASN1WithLabelMenu.add(openASN1WithLabelBuiltinDTDMenu);
        openASN1WithLabelMenu.add(openASN1WithLabelExternalDTDMenuItem);
        openASN1WithLabelExternalDTDMenuItem.addActionListener(this);

        openASN1WithLabelBuiltinDTDLDAPMenuItem = new JMenuItem("LDAP");
        openASN1WithLabelBuiltinDTDPKCS10MenuItem = new JMenuItem("PKCS#10");
        openASN1WithLabelBuiltinDTDPKCS7MenuItem = new JMenuItem("PKCS#7");
        openASN1WithLabelBuiltinDTDSNMPMenuItem = new JMenuItem("SNMP");
        openASN1WithLabelBuiltinDTDX509MenuItem = new JMenuItem("X.509");
        openASN1WithLabelBuiltinDTDMenu.add(openASN1WithLabelBuiltinDTDLDAPMenuItem);
        openASN1WithLabelBuiltinDTDMenu.add(openASN1WithLabelBuiltinDTDPKCS10MenuItem);
        openASN1WithLabelBuiltinDTDMenu.add(openASN1WithLabelBuiltinDTDPKCS7MenuItem);
        openASN1WithLabelBuiltinDTDMenu.add(openASN1WithLabelBuiltinDTDSNMPMenuItem);
        openASN1WithLabelBuiltinDTDMenu.add(openASN1WithLabelBuiltinDTDX509MenuItem);
        openASN1WithLabelBuiltinDTDLDAPMenuItem.addActionListener(this);
        openASN1WithLabelBuiltinDTDPKCS10MenuItem.addActionListener(this);
        openASN1WithLabelBuiltinDTDPKCS7MenuItem.addActionListener(this);
        openASN1WithLabelBuiltinDTDSNMPMenuItem.addActionListener(this);
        openASN1WithLabelBuiltinDTDX509MenuItem.addActionListener(this);

        saveAsXMLMenuItem = new JMenuItem("XML");
        saveAsASN1MenuItem = new JMenuItem("ASN.1");
        saveAsASN1ASDMenuItem = new JMenuItem("ASN.1 ASD");
        saveAsDTDMenuItem = new JMenuItem("DTD");
        saveAsMenu.add(saveAsXMLMenuItem);
        saveAsMenu.add(saveAsASN1MenuItem);
        saveAsMenu.addSeparator();
        saveAsMenu.add(saveAsASN1ASDMenuItem);
        saveAsMenu.add(saveAsDTDMenuItem);
        saveAsXMLMenuItem.addActionListener(this);
        saveAsASN1MenuItem.addActionListener(this);
        saveAsASN1ASDMenuItem.addActionListener(this);
        saveAsDTDMenuItem.addActionListener(this);

        cutMenuItem = new JMenuItem("Cut");
        copyMenuItem = new JMenuItem("Copy");
        pasteMenuItem = new JMenuItem("Paste");
        selectAllMenuItem = new JMenuItem("Select All");
        JMenu optionsMenu = new JMenu("Options");
        editMenu.add(cutMenuItem);
        editMenu.add(copyMenuItem);
        editMenu.add(pasteMenuItem);
        editMenu.add(selectAllMenuItem);
        editMenu.add(optionsMenu);
        cutMenuItem.addActionListener(this);
        copyMenuItem.addActionListener(this);
        pasteMenuItem.addActionListener(this);
        selectAllMenuItem.addActionListener(this);

        oidMenuItem = new JCheckBoxMenuItem("Named OID", true);
        cascadingPKCS7MenuItem = new JCheckBoxMenuItem("Cascading PKCS#7 Labeling", true);
        optionsMenu.add(oidMenuItem);
        optionsMenu.add(cascadingPKCS7MenuItem);

        aboutMenuItem = new JMenuItem("About");
        helpMenu.add(aboutMenuItem);
        aboutMenuItem.addActionListener(this);

        popupMenu = new JPopupMenu("XML Editor");
        
        aboutPopupMenuItem = new JMenuItem("About");
        exitPopupMenuItem = new JMenuItem("Exit");
        popupMenu.add(aboutPopupMenuItem);
        popupMenu.addSeparator();
        popupMenu.add(exitPopupMenuItem);
        aboutPopupMenuItem.addActionListener(this);
        exitPopupMenuItem.addActionListener(this);
        addMouseListener(new MouseAdapter() {
                public void mousePressed(MouseEvent me) {
                    if (me.getModifiers() == InputEvent.BUTTON3_MASK) {
                        popupMenu.show(me.getComponent(),me.getX(),me.getY());
                    }
                }
            });        

        addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent we) {
                    System.exit(0);
                }
            });
    }

    /**
      * main().
      */
    public static void main(String[] args) {
        String si = null;
        if (args.length != 0){
            if (args.length == 2) {
                if (args[0].equals("-sw")) si = args[1];
                else printUsage(-1);
            } else printUsage(-1);
        }

        XMLEditor editor = new XMLEditor("XML Editor", si);
        if (splashWindow != null) splashWindow.dispose(); 
        editor.setVisible(true);
    }

    /**
      * Invoked when an action occurs.
      */
    public void actionPerformed(ActionEvent ae) {
        // if selected the menu File -> Open -> XML
        if (ae.getSource() == openXMLMenuItem) {
            int state = fileChooser.showOpenDialog(this);
            if (state == JFileChooser.APPROVE_OPTION) {
                openFile = fileChooser.getSelectedFile();

                if (openFile != null) {
                    try{
                        docPanel.read(new FileReader(openFile), null);
                        docPanel.setFileName(openFile.toString());
                        logPanel.append("[XML file opened]: " + openFile.getName());
                        setEnabledAXMenuItem(true);
                        setEnabledADMenuItem(false);
                    } catch (FileNotFoundException fnfe) {
                        logPanel.append("[XML file open failed because of]: " + fnfe.toString());
                    } catch (IOException ioe) {
                        logPanel.append("[XML file open failed because of]: " + ioe.toString());
                    }
                }
            }

        // if selected the menu File -> Open -> ASN.1 -> Without Label
        } else if (ae.getSource() == openASN1WithoutLabelMenuItem) {
            int state = fileChooser.showOpenDialog(this);
            if (state == JFileChooser.APPROVE_OPTION) {
                openFile = fileChooser.getSelectedFile();

                if (openFile != null) {
                    try{
                        tran.setInputStream(new FileInputStream(openFile));
                        tran.setNamedOID(oidMenuItem.isSelected());
                        tran.startAS(docPanel.getWriter());
                        docPanel.flush();
                        docPanel.setFileName(openFile.toString());
                        logPanel.append("[ASN.1 file opened without label]: " + openFile.getName());
                        saveAsASN1MenuItem.setEnabled(false);
                        saveAsXMLMenuItem.setEnabled(true);
                        setEnabledADMenuItem(false);
                    } catch (FileNotFoundException fnfe) {
                        logPanel.append("[ASN.1 file open failed because of]: " + fnfe.toString());
                    } catch (SAXException saxe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + saxe.toString());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + ioe.toString());
                    }
                }
            }

        // if selected the menu File -> Open -> ASN.1 -> With Label -> Using External DTD
        } else if (ae.getSource() == openASN1WithLabelExternalDTDMenuItem) {
            dialog.showDialog();

            int pressedButton = dialog.getPressedButton();
            if (pressedButton == ExternalDTDDialog.OK_BUTTON) {
                openFile = new File(dialog.getASN1File());

                try{
                    tran.setInputStream(new FileInputStream(openFile));
                    tran.setDoctypeDeclaration(dialog.getRootName(), dialog.getDTD(), false);
                    tran.setNamedOID(oidMenuItem.isSelected());
                    tran.startASS(docPanel.getWriter());
                    docPanel.flush();
                    docPanel.setFileName(openFile.toString());
                    logPanel.append("[ASN.1 file opened with label using external DTD]: " + openFile.getName());
                    setEnabledAXMenuItem(true);
                    setEnabledADMenuItem(false);
                } catch (FileNotFoundException fnfe) {
                    logPanel.append("[ASN.1 file open failed because of]: " + fnfe.toString());
                } catch (SAXException saxe) {
                    logPanel.append("[ASN.1 file open failed because of]:" + saxe.toString());
                } catch (IOException ioe) {
                    logPanel.append("[ASN.1 file open failed because of]:" + ioe.toString());
                }
            }

        // if selected the menu File -> Open -> ASN.1 -> With Label -> Using Built-in DTD -> LDAP
        } else if (ae.getSource() == openASN1WithLabelBuiltinDTDLDAPMenuItem) {
            int state = fileChooser.showOpenDialog(this);
            if (state == JFileChooser.APPROVE_OPTION) {
                openFile = fileChooser.getSelectedFile();

                if (openFile != null) {
                    try{
                        tran.setInputStream(new FileInputStream(openFile));
                        tran.setDoctypeDeclaration(LDAP_ROOT_NAME, LDAP_DTD);
                        tran.setNamedOID(oidMenuItem.isSelected());
                        tran.startASS(docPanel.getWriter());
                        docPanel.flush();
                        docPanel.setFileName(openFile.toString());
                        logPanel.append("[ASN.1 file opened with label using built-in LDAP DTD]: " + openFile.getName());
                        setEnabledAXMenuItem(true);
                        setEnabledADMenuItem(false);
                    } catch (FileNotFoundException fnfe) {
                        logPanel.append("[ASN.1 file open failed because of]: " + fnfe.toString());
                    } catch (SAXException saxe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + saxe.toString());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + ioe.toString());
                    }
                }
            }

        // if selected the menu File -> Open -> ASN.1 -> With Label -> Using Built-in DTD -> PKCS#10
        } else if (ae.getSource() == openASN1WithLabelBuiltinDTDPKCS10MenuItem) {
            int state = fileChooser.showOpenDialog(this);
            if (state == JFileChooser.APPROVE_OPTION) {
                openFile = fileChooser.getSelectedFile();

                if (openFile != null) {
                    try{
                        tran.setInputStream(new FileInputStream(openFile));
                        tran.setDoctypeDeclaration(PKCS10_ROOT_NAME, PKCS10_DTD);
                        tran.setNamedOID(oidMenuItem.isSelected());
                        tran.startASS(docPanel.getWriter());
                        docPanel.flush();
                        docPanel.setFileName(openFile.toString());
                        logPanel.append("[ASN.1 file opened with label using built-in PKCS#10 DTD]: " + openFile.getName());
                        setEnabledAXMenuItem(true);
                        setEnabledADMenuItem(false);
                    } catch (FileNotFoundException fnfe) {
                        logPanel.append("[ASN.1 file open failed because of]: " + fnfe.toString());
                    } catch (SAXException saxe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + saxe.toString());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + ioe.toString());
                    }
                }
            }
        // if selected the menu File -> Open -> ASN.1 -> With Label -> Using Built-in DTD -> PKCS#7
        } else if (ae.getSource() == openASN1WithLabelBuiltinDTDPKCS7MenuItem) {
            int state = fileChooser.showOpenDialog(this);
            if (state == JFileChooser.APPROVE_OPTION) {
                openFile = fileChooser.getSelectedFile();

                if (openFile != null) {
                    try{
                        tran.setInputStream(new FileInputStream(openFile));
                        tran.setDoctypeDeclaration(PKCS7_ROOT_NAME, PKCS7_DTD);
                        tran.setNamedOID(oidMenuItem.isSelected());
                        if (cascadingPKCS7MenuItem.isSelected()) {
                            tran.startASSS(docPanel.getWriter());
                        }else {
                            tran.startASS(docPanel.getWriter());
                        }
                        docPanel.flush();
                        docPanel.setFileName(openFile.toString());
                        logPanel.append("[ASN.1 file opened with label using built-in PKCS#7 DTD]: " + openFile.getName());
                        if (cascadingPKCS7MenuItem.isSelected()) {
                            setEnabledAXMenuItem(true);
                        } else {
                            saveAsASN1MenuItem.setEnabled(false);
                            saveAsXMLMenuItem.setEnabled(true);
                        }
                        setEnabledADMenuItem(false);
                    } catch (FileNotFoundException fnfe) {
                        logPanel.append("[ASN.1 file open failed because of]: " + fnfe.toString());
                    } catch (SAXException saxe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + saxe.toString());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + ioe.toString());
                    }
                }
            }

        // if selected the menu File -> Open -> ASN.1 -> With Label -> Using Built-in DTD -> SNMP
        } else if (ae.getSource() == openASN1WithLabelBuiltinDTDSNMPMenuItem) {
            int state = fileChooser.showOpenDialog(this);
            if (state == JFileChooser.APPROVE_OPTION) {
                openFile = fileChooser.getSelectedFile();

                if (openFile != null) {
                    try{
                        tran.setInputStream(new FileInputStream(openFile));
                        tran.setDoctypeDeclaration(SNMP_ROOT_NAME, SNMP_DTD);
                        tran.setNamedOID(oidMenuItem.isSelected());
                        tran.startASS(docPanel.getWriter());
                        docPanel.flush();
                        docPanel.setFileName(openFile.toString());
                        logPanel.append("[ASN.1 file opened with label using built-in SNMP DTD]: " + openFile.getName());
                        setEnabledAXMenuItem(true);
                        setEnabledADMenuItem(false);
                    } catch (FileNotFoundException fnfe) {
                        logPanel.append("[ASN.1 file open failed because of]: " + fnfe.toString());
                    } catch (SAXException saxe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + saxe.toString());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + ioe.toString());
                    }
                }
            }

        // if selected the menu File -> Open -> ASN.1 -> With Label -> Using Built-in DTD -> X.509
        } else if (ae.getSource() == openASN1WithLabelBuiltinDTDX509MenuItem) {
            int state = fileChooser.showOpenDialog(this);
            if (state == JFileChooser.APPROVE_OPTION) {
                openFile = fileChooser.getSelectedFile();

                if (openFile != null) {
                    try{
                        tran.setInputStream(new FileInputStream(openFile));
                        tran.setDoctypeDeclaration(X509_ROOT_NAME, X509_DTD);
                        tran.setNamedOID(oidMenuItem.isSelected());
                        tran.startASS(docPanel.getWriter());
                        docPanel.flush();
                        docPanel.setFileName(openFile.toString());
                        logPanel.append("[ASN.1 file opened with label using X.509 DTD]: " + openFile.getName());
                        setEnabledAXMenuItem(true);
                        setEnabledADMenuItem(false);
                    } catch (FileNotFoundException fnfe) {
                        logPanel.append("[ASN.1 file open failed because of]: " + fnfe.toString());
                    } catch (SAXException saxe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + saxe.toString());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 file open failed because of]:" + ioe.toString());
                    }
                }
            }

        // if selected the menu File -> Open -> ASN.1 ASD
        } else if (ae.getSource() == openASN1ASDMenuItem) {
            int state = fileChooser.showOpenDialog(this);
            if (state == JFileChooser.APPROVE_OPTION) {
                openFile = fileChooser.getSelectedFile();

                if (openFile != null) {
                    try{
                        docPanel.read(new FileReader(openFile), null);
                        docPanel.setFileName(openFile.toString());
                        logPanel.append("[ASN.1 ASD file opened]: " + openFile.getName());
                        setEnabledAXMenuItem(false);
                        setEnabledADMenuItem(true);
                    } catch (FileNotFoundException fnfe) {
                        logPanel.append("[ASN.1 ASD file open failed because of]: " + fnfe.toString());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 ASD file open failed because of]: " + ioe.toString());
                    }
                }
            }

        // if selected the menu File -> Save
        } else if (ae.getSource() == saveMenuItem) {
            if (!docPanel.getText().equals("")) {
                if (openFile != null) {
                    try{
                        FileWriter fw = new FileWriter(openFile);
                        docPanel.write(fw);
                        fw.close();
                        logPanel.append("[File saved]: " + openFile.getName());
                    } catch (IOException ioe) {
                        logPanel.append("[File save failed because of]: " + ioe.toString());
                    }
                } else {
                    int state = fileChooser.showSaveDialog(this);
                    if (state == JFileChooser.APPROVE_OPTION) {            
                        File openFile = fileChooser.getSelectedFile();

                        try{
                            FileWriter fw = new FileWriter(openFile);
                            docPanel.write(fw);
                            fw.close();
                            docPanel.setFileName(openFile.toString());
                            logPanel.append("[File saved]: " + openFile.getName());
                        } catch (IOException ioe) {
                            logPanel.append("[File save failed because of]: " + ioe.toString());
                        }
                    }
                }
            } else {
                logPanel.append("[File save failed because of]: File contains no data");
            }

        // if selected the menu File -> Save As -> XML
        } else if (ae.getSource() == saveAsXMLMenuItem) {
            if (!docPanel.getText().equals("")) {
                int state = fileChooser.showSaveDialog(this);
                if (state == JFileChooser.APPROVE_OPTION) {            
                    File saveFile = fileChooser.getSelectedFile();

                    try{
                        FileWriter fw = new FileWriter(saveFile);
                        docPanel.write(fw);
                        fw.close();
                        logPanel.append("[XML file saved as]: " + saveFile.getName());
                    } catch (IOException ioe) {
                        logPanel.append("[XML file save failed because of]: " + ioe.toString());
                    }
                }
            } else {
                logPanel.append("[XML file save failed because of]: File contains no data");
            }

        // if selected the menu File -> Save As -> ASN.1
        } else if (ae.getSource() == saveAsASN1MenuItem) {
            if (!docPanel.getText().equals("")) {
                int state = fileChooser.showSaveDialog(this);
                if (state == JFileChooser.APPROVE_OPTION) {            
                    File saveFile = fileChooser.getSelectedFile();

                    try{
                        tran.setReader(docPanel.getReader());
                        FileOutputStream fos = new FileOutputStream(saveFile);
                        tran.startSSA(fos);
                        fos.close();
                        logPanel.append("[ASN.1 file saved as]: " + saveFile.getName());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 file save failed because of]: " + ioe.toString());
                    } catch (SAXException saxe) {
                        logPanel.append("[ASN.1 file save failed because of]: " + saxe.toString());
                    }
                }
            } else {
                logPanel.append("[ASN.1 file save failed because of]: File contains no data");
            }

        // if selected the menu File -> Save As -> ASN.1 ASD
        } else if (ae.getSource() == saveAsASN1ASDMenuItem) {
            if (!docPanel.getText().equals("")) {
                int state = fileChooser.showSaveDialog(this);
                if (state == JFileChooser.APPROVE_OPTION) {            
                    File saveFile = fileChooser.getSelectedFile();

                    try{
                        FileWriter fw = new FileWriter(saveFile);
                        docPanel.write(fw);
                        fw.close();
                        logPanel.append("[ASN.1 ASD file saved as]: " + saveFile.getName());
                    } catch (IOException ioe) {
                        logPanel.append("[ASN.1 ASD file save failed because of]: " + ioe.toString());
                    }
                }
            } else {
                logPanel.append("[ASN.1 ASD file save failed because of]: File contains no data");
            }

        // if selected the menu File -> Save As -> DTD
        } else if (ae.getSource() == saveAsDTDMenuItem) {
            if (!docPanel.getText().equals("")) {
                int state = fileChooser.showSaveDialog(this);
                if (state == JFileChooser.APPROVE_OPTION) {            
                    File saveFile = fileChooser.getSelectedFile();

                    try{
                        tran.setReader(docPanel.getReader());
                        FileWriter fw = new FileWriter(saveFile);
                        tran.startAD(fw);
                        fw.close();
                        logPanel.append("[DTD file saved as]: " + saveFile.getName());
                    } catch (IOException ioe) {
                        logPanel.append("[DTD file save failed because of]: " + ioe.toString());
                    } catch (ParseException pe) {
		        logPanel.append("[DTD file save failed because of]: " + pe.toString());
		    }
                }
            } else {
                logPanel.append("[DTD file save failed because of]: File contains no data");
            }

        // if selected the menu File -> Exit
        } else if (ae.getSource() == exitMenuItem) {
            System.exit(0);

        // if selected the menu Edit -> Cut
        } else if (ae.getSource() == cutMenuItem) {
            docPanel.cut();

        // if selected the menu Edit -> Copy
        } else if (ae.getSource() == copyMenuItem) {
            docPanel.copy();

        // if selected the menu Edit -> Paste
        } else if (ae.getSource() == pasteMenuItem) {
            docPanel.paste();

        // if selected the menu Edit -> SelectAll
        } else if (ae.getSource() == selectAllMenuItem) {
            docPanel.selectAll();

        // if selected the menu Help -> About
        } else if (ae.getSource() == aboutMenuItem) {
            JOptionPane.showMessageDialog(this,ABOUT_MESSAGE,"About XML Editor",JOptionPane.INFORMATION_MESSAGE);

        // if selected the popup menu About
        } else if (ae.getSource() == aboutPopupMenuItem) {
            repaint();
            JOptionPane.showMessageDialog(this,ABOUT_MESSAGE,"About XML Editor",JOptionPane.INFORMATION_MESSAGE);

        // if selected the popup menu Exit
        } else if (ae.getSource() == exitPopupMenuItem) {
            System.exit(0);
        }
    }

    /**
      * Prints the usage message and exits this program with a status.
      *
      * @param <CODE>stat</CODE> - a status of this program.
      */
    private static void printUsage(int stat) {
        System.err.println("Usage: java asn1.XMLEditor -sw <image file>");
        System.exit(stat);
    }

    /**
      * Returns the file descriptor of the given file.
      *
      * @param <CODE>file</CODE> - the name of the file.
      *
      * @return a file descriptor.
      */
    private String getFileDescriptor(String file) {
        if (file == null) return "";

        // Gets the file descriptor
        StringTokenizer st = new StringTokenizer(file, ".");
        String fd = "";
        while(st.hasMoreElements()) fd = st.nextToken();

        return fd;
    }

    /**
      * Enables or disables the menu item associated with to ASN.1/XML.
      *
      * @param <CODE>b</CODE> - true to enable the item, false to disable.
      */
    private void setEnabledAXMenuItem(boolean b) {
        saveAsASN1MenuItem.setEnabled(b);
        saveAsXMLMenuItem.setEnabled(b);
    }

    /**
      * Enables or disables the menu item associated with ASN.1 ASD/DTD.
      *
      * @param <CODE>b</CODE> - true to enable the item, false to disable.
      */
    private void setEnabledADMenuItem(boolean b) {
        saveAsASN1ASDMenuItem.setEnabled(b);
        saveAsDTDMenuItem.setEnabled(b);
    }
}
