/*
 * (C) Copyright IBM Corp. 1999,2000,2001,2002  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package xss4jsamples.dsig;

import com.ibm.dom.util.StandardErrorHandler;
import com.ibm.xml.dsig.Canonicalizer;
import com.ibm.xml.dsig.IDResolver;
import com.ibm.xml.dsig.KeyInfo;
import com.ibm.xml.dsig.Reference;
import com.ibm.xml.dsig.SignatureContext;
import com.ibm.xml.dsig.SignatureMethod;
import com.ibm.xml.dsig.SignatureStructureException;
import com.ibm.xml.dsig.TemplateGenerator;
import com.ibm.xml.dsig.Transform;
import com.ibm.xml.dsig.XSignature;
import com.ibm.xml.dsig.XSignatureException;
import com.ibm.xml.dsig.util.DOMParserNS;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Hashtable;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * Sample signer.
 *
 * @version $Id: SampleSign2.java,v 1.11 2004/04/14 06:41:16 kent Exp $
 * @author TAMURA Kent &lt;kent@trl.ibm.co.jp&gt;
 */
public class SampleSign2 {
    static void setupReferenceExternal(TemplateGenerator signature, String uri)
            throws IOException {
        Reference ref = signature.createReference(uri);
        signature.addReference(ref);
    }

    static void setupReferenceEmbededXML(TemplateGenerator signature,
                                         String uri, IDGenerator idg)
            throws IOException, ParserConfigurationException, SAXException {
        DocumentBuilder builder = DOMParserNS.createBuilder();
        builder.setErrorHandler(new StandardErrorHandler());
        Document doc = builder.parse(uri);
        String id = idg.generate();
        Element object = signature.wrapWithObject(doc, id);
        idg.register(id, object);
        Reference ref = signature.createReference(object);
        ref.addTransform(Transform.W3CC14N2);
        signature.addReference(ref);
    }

    public static void main(String[] argv)
            throws XSignatureException, IOException, KeyStoreException,
            SignatureStructureException, NoSuchAlgorithmException,
            CertificateException, UnrecoverableKeyException,
            ParserConfigurationException, SAXException {
        String keystorepath = xss4jsamples.dsig.SignatureUtil.getDefaultKeystoreName();
        String alias = null;
        char[] storepass = null;
        char[] keypass = null;
        String prefix = null;
        int objectIndex = -1;
        int indent = 2;

        System.err.println("Key store: "+keystorepath);
        for (int i = 0;  i < argv.length;  i++) {
            if (argv[i].equals("-prefix")) {
                prefix = argv[++i];
            } else if (argv[i].equals("-indent")) {
                indent = Integer.parseInt(argv[++i]);
            } else {
                if (alias == null)
                    alias = argv[i];
                else if (storepass == null)
                    storepass = argv[i].toCharArray();
                else if (keypass == null)
                    keypass = argv[i].toCharArray();
                else {
                    objectIndex = i;
                    break;
                }
            }
        }
        if (objectIndex < 0) {
            System.err.println("");
            System.err.println("Usage: dsig.SampleSign2 <signer-alias> "
                               +"<storepass> <keypass> <resource> <resource> ...");
            System.err.println("");
            System.err.println("<resource>:  -ext http://.... : External resource.");
            System.err.println("             -embxml http://.... : Embeded XML resource.");
            System.exit(1);
        }

        //
        // Initialize
        //
        Document doc;
        IDGenerator idg;
        TemplateGenerator signatureGen;

        doc = DOMParserNS.createBuilder().newDocument();
        idg = new IDGenerator(doc, "Res");
        signatureGen = new TemplateGenerator(doc, XSignature.SHA1,
                                             Canonicalizer.W3C2,
                                             SignatureMethod.DSA);
        signatureGen.setPrefix(prefix);
        if (indent < 0) {
            signatureGen.setIndentation(false);
        } else {
            signatureGen.setIndentation(true);
            signatureGen.setUnit(indent);
        }

        //
        // Setup ReferenceObjects
        //
        for (int i = objectIndex;  i < argv.length;  i += 2) {
            if (argv[i].equals("-ext")) {
                if (i+1 >= argv.length)
                    throw new IllegalArgumentException("An -ext option requires one URL.");
                setupReferenceExternal(signatureGen, argv[i+1]);
            } else if (argv[i].equals("-embxml")) {
                if (i+1 >= argv.length)
                    throw new IllegalArgumentException("An -embxml option requires one URL.");
                setupReferenceEmbededXML(signatureGen, argv[i+1], idg);
            } else {
                System.err.println("Warning: Unknown option: "+argv[i]);
            }
        }

        //
        // Setup a certificate and a private key.
        //
        KeyStore keystore = KeyStore.getInstance("JKS");
        keystore.load(new FileInputStream(keystorepath), storepass);
        X509Certificate cert = (X509Certificate)keystore.getCertificate(alias);
        Key key = keystore.getKey(alias, keypass); // a private key
        if (key == null) {
            System.err.println("Could not get a key: "+alias);
            System.exit(1);
        }
        KeyInfo keyInfo = xss4jsamples.dsig.SignatureUtil.createKeyInfo(cert);
        Element signatureElement = signatureGen.getSignatureElement();
        keyInfo.insertTo(signatureElement, prefix, signatureGen);

        //
        // Sign
        //
        long startSign = System.currentTimeMillis();
        SignatureContext sigContext = new SignatureContext();
        sigContext.setIDResolver(idg);
        sigContext.sign(signatureElement, key);
        long endSign = System.currentTimeMillis();
        System.err.println("Sign: "+(endSign-startSign)+"ms");

        //
        // Output
        //
        doc.appendChild(signatureElement);
        xss4jsamples.dsig.SignatureUtil.printDocument(doc, System.out);
    }

    static class IDGenerator implements IDResolver {
        Document doc;
        int counter;
        String base;
        Hashtable hash;
        IDGenerator(Document doc, String base) {
            this.doc = doc;
            this.counter = 0;
            this.base = base;
            this.hash = new Hashtable();
        }
        String generate() {
            return this.base+this.counter++;
        }
        void register(String id, Element el) {
            this.hash.put(id, el);
        }
        public Element resolveID(Document doc, String id) {
            if (doc != this.doc)  return null;
            return (Element)this.hash.get(id);
        }
    }
}
