/*
 * (C) Copyright IBM Corp. 2000,2001  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package xss4jsamples.dsig;

import com.ibm.dom.util.StandardErrorHandler;
import com.ibm.xml.dsig.KeyInfo;
import com.ibm.xml.dsig.ResourceShower;
import com.ibm.xml.dsig.SignatureContext;
import com.ibm.xml.dsig.Validity;
import com.ibm.xml.dsig.XSignature;
import com.ibm.xml.dsig.util.AdHocIDResolver;
import com.ibm.xml.dsig.util.DOMParserNS;
import com.ibm.xml.dsig.util.SimpleKey;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.Key;
import java.security.cert.CertificateException;
import java.security.cert.X509CRL;
import java.security.cert.X509Certificate;
import java.util.Set;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;                    // DOM Level 2!
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 *
 * @version $Id: VerifyGUI.java,v 1.6 2004/04/14 06:44:35 kent Exp $
 * @author TAMURA Kent &lt;kent@trl.ibm.co.jp&gt;
 */
public class VerifyGUI {

    static class TableModelImpl extends AbstractTableModel {
        static final String[] ids = {"Location", "Status"};
        Validity validity;

        TableModelImpl(Validity v) {
            this.validity = v;
        }

        public int getRowCount() {
            return this.validity.getNumberOfReferences();
        }
        public int getColumnCount() {
            return 2;
        }
        public String getColumnName(int column) {
            return ids[column];
        }
        public Object getValueAt(int row, int column) {
            switch (column) {
              case 0:
                return this.validity.getReferenceURI(row);
              case 1:
                boolean validp = this.validity.getReferenceValidity(row);
                return validp ? "OK" : "NG: "+this.validity.getReferenceMessage(row);
            }
            return null;
        }
  
    }

    private static final String WINDOW_TITLE = "XML-Siganture Verifier";

    public static void main(String[] argv) {
        InputStream is = null;
        int saveno = -2;
        String savename = null;
        Key key = null;
        //
        // Analyze the command line
        //
        for (int i = 0;  i < argv.length;  i ++) {
            if (argv[i].equals("-s")) {
                saveno = Integer.parseInt(argv[i+1]);
                savename = argv[i+2];
                i += 2;
            } else if (argv[i].equals("-keyfile")) {
                File keyfile = new File(argv[i+1]);
                int length = (int)keyfile.length();
                System.err.print("Reading a key file ...");
                byte[] rawkey = new byte[length];
                try {
                    FileInputStream fis = new FileInputStream(keyfile);
                    fis.read(rawkey);
                    fis.close();
                    System.err.println(" done: "+(length*8)+" bit");
                    key = new SimpleKey(rawkey);
                } catch (IOException ioe) {
                    System.err.println(ioe);
                }
                i++;
            } else {
                try {
                    is = new FileInputStream(argv[i]);
                    break;
                } catch (IOException ioe) {
                    System.err.println(ioe);
                }
            }
        }
        if (is == null)  is = System.in;

        JFrame frame = new JFrame("Verifier for W3C/IETF XML-Signature");
        frame.addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    System.exit(0);
                }
            });
        final int FRAMEWIDTH = 480, FRAMEHEIGHT = 320;
        Dimension screen = frame.getToolkit().getScreenSize();
        frame.setBounds((screen.width-FRAMEWIDTH)/2, (screen.height-FRAMEHEIGHT)/2,
                        FRAMEWIDTH, FRAMEHEIGHT);
        Container cont = frame.getContentPane();
        frame.show();

        try {
            //
            // Parses a signature document.
            //
            InputSource isrc = new InputSource(is);
            //isrc.setEncoding("UTF-8");
            Document doc = parse(isrc);

            //
            // Searches for a Signature element
            //
            NodeList nodeList = doc.getElementsByTagNameNS(XSignature.XMLDSIG_NAMESPACE,
                                                           "Signature");
            if (nodeList.getLength() == 0) {
                JOptionPane.showMessageDialog(null, "The specified document has no Siganture element.",
                                              WINDOW_TITLE, JOptionPane.ERROR_MESSAGE);
                System.exit(1);
            }
            Element signature = (Element)nodeList.item(0);
            
            SignatureContext sigContext = new SignatureContext();
            sigContext.setIDResolver(new AdHocIDResolver(doc));
            sigContext.setResourceShower(saveno <= -2 ? null : new ShowerImpl(saveno, savename));

            Element keyInfoElement = KeyInfo.searchForKeyInfo(signature);
            Validity validity = key != null
                ? verify(sigContext, signature, key, true)
                : verify(sigContext, signature, keyInfoElement, true);

            //
            // Creates result view
            //
            JPanel panel = new JPanel();
            panel.setLayout(new GridLayout(2, 1));
            JLabel lab = new JLabel(validity.getCoreValidity() ? "Core Validity: OK"
                                    : "Core Validity: NG");
            lab.setFont(new Font("Serif", Font.PLAIN, 30));
            panel.add(lab);
            String signedInfoMes = validity.getSignedInfoValidity()
                ? "Signature Validity: OK"
                : "Signature Validity: NG ("+validity.getSignedInfoMessage()+")";
            lab = new JLabel(signedInfoMes);
            lab.setFont(new Font("Serif", Font.PLAIN, 25));
            panel.add(lab);
            cont.add(panel, "North");
            TableModelImpl impl = new TableModelImpl(validity);
            JTable table = new JTable(impl);
            table.setFont(new Font("Serif", Font.PLAIN, 20));

            cont.add(new JScrollPane(table));
            cont.validate();
            cont.repaint();

        } catch (SAXException saxe) {
            Exception ex = saxe.getException();
            if (ex == null)  ex = saxe;
            JOptionPane.showMessageDialog(null, ex, WINDOW_TITLE, JOptionPane.ERROR_MESSAGE);
            System.exit(1);
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(null, ex, WINDOW_TITLE, JOptionPane.ERROR_MESSAGE);
            System.exit(1);
        }
    }

    static Validity verify(SignatureContext sigContext, Element signature, Key key, boolean verbose) {
        //
        // Checks validity of the Signature element
        //
        long startTime = System.currentTimeMillis();
        Validity validity = sigContext.verify(signature, key);
        long elapsedTime = System.currentTimeMillis()-startTime;
        if (verbose)
            System.err.println("Time to verify: "+elapsedTime+" [msec]");
        return validity;
    }
    static Validity verify(SignatureContext sigContext, Element signature,
                           Element keyInfoElement, boolean verbose) {
        if (keyInfoElement == null) {
            return null;
        }
        Key key = null;
        KeyInfo keyInfo;
        try {
            keyInfo = new KeyInfo(keyInfoElement);
        } catch (Exception ex) {
            ex.printStackTrace();
            return null;
        }

        Key keyInKeyValue = keyInfo.getKeyValue();
        if (keyInKeyValue != null) {
            if (verbose)
                System.err.println("The signature has a KeyValue element.");
            if (key == null)
                key = keyInKeyValue;
        }

        KeyInfo.X509Data[] x5data = keyInfo.getX509Data();
        if (x5data != null && x5data.length > 0) {
            if (verbose)
                System.err.println("The signature has one or more X509Data elements.");
            for (int i = 0;  i < x5data.length;  i++) {
                Key xkey = printX509DataInfo(x5data[i], verbose);
                if (key == null)
                    key = xkey;
            }
        }

        Element[] retrievals = keyInfo.getRetrievalMethods();
        if (retrievals != null && retrievals.length > 0) {
            if (verbose)
                System.err.println("The signature has one or more RetrievalMethod elements.");
            for (int i = 0;  i < retrievals.length;  i++) {
                String type = retrievals[i].getAttribute("Type");
                System.err.println("Type: "+type);
                if (type.equals(KeyInfo.X509DATA)) {
                    try {
                        KeyInfo.X509Data x5d = (KeyInfo.X509Data)sigContext.retrieve(retrievals[i]);
                        Key xkey = printX509DataInfo(x5d, verbose);
                        if (key == null) {
                            System.err.println("Uses a key from a RetrievalMethod.");
                            key = xkey;
                        }
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
            }
        }
        return verify(sigContext, signature, key, verbose);
    }

    /**
     * Print information in an X509Data and get a key.
     */
    static Key printX509DataInfo(KeyInfo.X509Data x5data, boolean verbose) {
        if (verbose)
            System.err.println("Checks an X509Data:");
        Key key = null;
        X509CRL crl = x5data.getCRL();
        if (crl != null) {
            if (verbose) {
                System.err.println("CRL Information:");
                System.err.println("\tVersion: "+crl.getVersion());
                System.err.println("\tIssuer: "+crl.getIssuerDN());
                System.err.println("\tUpdated: "+crl.getThisUpdate());
                System.err.println("\tNext update: "+crl.getNextUpdate());
                Set certs = crl.getRevokedCertificates();
                System.err.println("\tNumber of revoked certs: "
                                   +(certs != null ? certs.size() : 0));
            }
        }
        X509Certificate[] certs = x5data.getCertificates();
        if (certs == null || certs.length <= 0) {
            System.err.println("No certificates.");
            return null;
        }
        if (verbose)
            System.err.println(certs.length+" certificate(s).");
        for (int j = 0;  j < certs.length;  j++) {
            boolean revoked = false;
            X509Certificate cert = certs[j];
            if (verbose) {
                System.err.println("");
                System.err.println("Certificate Information:");
                System.err.println("\tVersion: "+cert.getVersion());
            }
            String validity = "OK";
            try {
                cert.checkValidity();
            } catch (CertificateException certex) {
                validity = certex.getMessage();
            }
            if (verbose) {
                System.err.println("\tValidity: "+validity);
                System.err.println("\tSubjectDN: "+cert.getSubjectDN());
                System.err.println("\tIssuerDN: "+cert.getIssuerDN());
                System.err.println("\tSerial#: 0x"+cert.getSerialNumber().toString(16));
            }
            if (crl != null && crl.getIssuerDN().equals(cert.getIssuerDN())) {
                if (crl.getRevokedCertificate(cert.getSerialNumber()) != null) {
                    System.err.println("\tThis certificate was revoked!!!");
                    JOptionPane.showMessageDialog(null, "The certificate was revoked.",
                                                  WINDOW_TITLE,
                                                  JOptionPane.INFORMATION_MESSAGE);
                    revoked = true;
                }
            }
            /* XXX: This is not correct if we have a chain. */
            if (!revoked && key == null)
                key = cert.getPublicKey();
        }
        return key;
    }

    static class ShowerImpl implements ResourceShower {
        int number;
        String name;
        ShowerImpl(int no, String name) {
            this.number = no;
            this.name = name;
        }
        public void showSignedResource(Element owner, int count, String uri, String type,
                                       byte[] content, String encoding) {
            if (this.number == count) {
                try {
                    OutputStream os = new FileOutputStream(this.name);
                    os.write(content);
                    os.close();
                } catch (IOException ioe) {
                    ioe.printStackTrace();
                }
            }
        }
    }

    public static Document parse(InputSource src)
            throws IOException, SAXException, ParserConfigurationException {
        DocumentBuilder builder = DOMParserNS.createBuilder();
        //builder.setErrorHandler(...);
        //XPathUtil.fixTree(doc);
        return builder.parse(src);
    }
}

